#include "config.h"
#include "player.h"

using namespace std;

void initWorkers(Player* p1, Player* p2, bool print = true);

int Board[BOARD_SIZE][BOARD_SIZE];
Worker W[NUM_WORKERS];

const string ANN1File = "ANNL.txt";
const string ANN2File = ANN1File;


int main() {
	int choice = 0;
	cout << "Choose player type for player 1: " << endl <<
		"1. Human" << endl <<
		"2. Random" << endl <<
		"3. Heuristic" << endl <<
		"4. ANN" << endl;

	while (true) {
		cin >> choice;
		if (cin.fail()) { //checks to see if choice is not a int
			cout << "Invalid choice." << endl;
			cin.clear();
			cin.ignore(INT64_MAX, '\n');
		}
		else {
			if (!(choice > 0 && choice < 5)) { //ensures that selection is within range
				cout << "Invalid choice." << endl;
			}
			else {
				break;
			}
		}
	}

	Player* p1;
	if (choice == 1) {
		p1 = new plyrHuman("Human 1", RED);
	}
	else if (choice == 2) {
		p1 = new plyrRandom("Random 1", PLAYER1_COLOR);
	}
	else if (choice == 3) {
		p1 = new plyrHeuristic("Heuristic 1", PLAYER1_COLOR);
	}
	else {
		p1 = new plyrANN(ANN1File, "ANN 1", PLAYER1_COLOR);
	}

	choice = 0;
	cout << "Choose player type for player 2: " << endl <<
		"1. Human" << endl <<
		"2. Random" << endl <<
		"3. Heuristic" << endl <<
		"4. ANN" << endl;

	while (true) {
		cin >> choice;
		if (cin.fail()) { //checks to see if choice is not a int
			cout << "Invalid choice." << endl;
			cin.clear();
			cin.ignore(INT64_MAX, '\n');
		}
		else {
			if (!(choice > 0 && choice < 5)) { //ensures that selection is within range
				cout << "Invalid choice." << endl;
			}
			else {
				break;
			}
		}
	}

	Player* p2;

	if (choice == 1) {
		p2 = new plyrHuman("Human 2", PLAYER2_COLOR);
	}
	else if (choice == 2) {
		p2 = new plyrRandom("Random 2", PLAYER2_COLOR);
	}
	else if (choice == 3) {
		p2 = new plyrHeuristic("Heuristic 2", PLAYER2_COLOR);
	}
	else {
		p2 = new plyrANN(ANN2File, "ANN 2", PLAYER2_COLOR);
	}

	double v[50];
	double gameTree[128][50];


	initWorkers(p1, p2);

	cout << endl << endl << endl;
	cout << "GAME START" << endl << endl;

	int result = 0;
	int turn = 0;
	bool gameLoop = true;

	srand(time(NULL));

	while (gameLoop) {
		turn = (turn + 1) % 2;
		printBoard();

		if (turn == 0) {
			
			result = p1->takeTurn();
		}
		else {
			p2->printTransitions();
			//extractBoardState(p2->getColor(), v);
			
			result = p2->takeTurn();
		}

		if (result != 0) {
			printBoard();
			cout << "Game Over!" << endl;
			if (result == 1) {
				if (turn == 0) {
					cout << p1->getName() << " wins!" << endl;
				}
				else {
					cout << p2->getName() << " wins!" << endl;
				}
			}
			else {
				if (turn == 0) {
					cout << p2->getName() << " wins!" << endl;
				}
				else {
					cout << p1->getName() << " wins!" << endl;
				}
			}

			break;
		}
	}
		
	/*
	else if (option == 'H') {
		ifstream fin;
		fin.open("saveboards.txt");
		plyrHeuristic* p1 = new plyrHeuristic("Player 1", RED);
		plyrHeuristic* p2 = new plyrHeuristic("Player 2", GREEN);
		int turn = 0;

		int original_color = 0;
		double original_board[50];

		double boardstates[2000][50];
		double measuredvalues[2000];
		int numstates = 0;

		while (!fin.eof()) {
			importBoardState(original_color, p1, p2, fin);
			if (original_color == 0) {
				extractBoardState(p1->getColor(), original_board);
			}
			else {
				extractBoardState(p2->getColor(), original_board);

			}

			for (int i = 0; i < 50; i++) {
				boardstates[numstates][i] = original_board[i];
			}
			//printVector(boardstates[numstates]);
			numstates++;
		}

		fin.close();

		srand(4);
		int index = rand() % numstates;
		importBoardState(turn, p1, p2, boardstates[index]);

		double current[50];

		for (int i = 0; i < 10; i++) {
			cout << "TURN " << i << endl;
			turn = (turn + 1) % 2;
			extractBoardState(turn, current);
			printVector(current);
			printBoard();
			if (turn == 0) {
				cout << "Value (GREEN): " << p1->getEval(current) << endl << endl;

				p2->takeTurn();
			}
			else {
				cout << "Value (RED): " << p2->getEval(current) << endl << endl;

				p1->takeTurn();
			}
			//p1->printTransitions();
		}

		delete p1;
		delete p2;
	}



	else if (option == 'Z') {
		ifstream fin;
		fin.open("saveboards.txt");
		Player* p1;
		Player* p2;
		p1 = new plyrHeuristic("Player 1", RED);
		p2 = new plyrHeuristic("Player 2", GREEN);

		int original_color = 0;
		double board[50];

		double boardstates[2000][50];
		double measuredvalues[2000];
		int numstates = 0;
		int numvalues = 0;

		while (!fin.eof()) {
			importBoardState(original_color, p1, p2, fin);
			if (original_color == 0) {
				extractBoardState(p1->getColor(), board);
			}
			else {
				extractBoardState(p2->getColor(), board);

			}

			for (int i = 0; i < 50; i++) {
				boardstates[numstates][i] = board[i];
			}
			//printVector(boardstates[numstates]);
			numstates++;
		}
		fin.close();

		double temp;
		fin.open("savevalues.txt");
		while (!fin.eof()) {
			fin >> measuredvalues[numvalues];
			numvalues++;
		}
		fin.close();

		double** gradient;
		gradient = new double* [500];
		for (int i = 0; i < 500; i++) {
			gradient[i] = new double[7850];
			for (int j = 0; j < 7850; j++) {
				gradient[i][j] = 0;
			}
		}

		cout << "begin testing" << endl;
		double result = 0;

		cout << "Boardstate 0" << endl;
		int turn = p1->getColor();
		importBoardState(turn, p1, p2, boardstates[4]);
		printVector(boardstates[4]);

		plyrANN* p3;
		p3 = new plyrANN("ANN0.txt", "Player 1", RED);

		for (int i = 0; i < 50; i++) {
			result = p3->getLogit(boardstates[4], 1, i);
			if (result < 0)
				setColor(RED);
			else
				setColor(WHITE);
			cout << setw(3) << result;
		}
		cout << endl;
		for (int i = 0; i < 50; i++) {
			cout << setw(3) << p3->getHidden(boardstates[4], 1, i);
		}
		cout << endl;


		for (int i = 0; i < 50; i++) {
			result = p3->getLogit(boardstates[4], 2, i);
			if (result < 0)
				setColor(RED);
			else
				setColor(WHITE);
			cout << setw(3) << result;
		}
		cout << endl;
		
		for (int i = 0; i < 50; i++) {
			cout << setw(3) << p3->getHidden(boardstates[4], 2, i);
		}
		cout << endl;

		for (int i = 0; i < 50; i++) {
			result = p3->getLogit(boardstates[4], 3, i);
			if (result < 0)
				setColor(RED);
			else
				setColor(WHITE);
			cout << setprecision(2) << setw(3) << result;
		}
		cout << endl;
		for (int i = 0; i < 50; i++) {
			cout << setprecision(2) << setw(3) << p3->getHidden(boardstates[4], 3, i);
		}
		cout << endl;


		cout << p3->getEval(boardstates[4]) << endl;

	}
	*/

	return 0;
}
/// <summary>
/// Initializes all 4 workers at game start.
/// </summary>
/// <param name="p1">Player 1</param>
/// <param name="p2">Player 2</param>
/// <param name="print">Whether or not to printboard between selections</param>
void initWorkers(Player* p1, Player* p2, bool print) {
	int turn = rand() % 2;

	for (int i = 0; i < NUM_WORKERS; i++) {
		if (print == true)
			printBoard();
		turn = (turn + 1) % 2;
		if (turn == 0) {
			p1->setWorker(i);
		}
		else {
			p2->setWorker(i);
		}
	}
	return;
}