#include "player.h"


//Player class functions
/// <summary>
/// Accessor for color.
/// </summary>
/// <returns>this.color</returns>
int Player::getColor() const {
    return color;
}
/// <summary>
/// Accessor for name.
/// </summary>
/// <returns>this.name</returns>
string Player::getName() const {
	return name;
}
/// <summary>
/// Mutator for color.
/// </summary>
/// <param name="col">Updated value for this.color</param>
void Player::setColor(int col) {
	color = col;
}
/// <summary>
/// Mutator for name.
/// </summary>
/// <param name="nm">Updated value for this.name</param>
void Player::setName(string nm) {
	name = nm;
}
/// <summary>
/// Populates m with indices in W of this players workers.
/// </summary>
/// <param name="m">Array to store workers</param>
void Player::getWorkers(int m[NUM_WORKERS_PER_PLAYER]) {
    int index = 0;
    for (int i = 0; i < NUM_WORKERS; i++) {
        if (W[i].color == color) {
            m[index] = i;
            index++;
        }
    }
    return;
}
/// <summary>
/// Populates m[workers][movements][placements] with valid moves in a single turn.
/// </summary>
/// <param name="m">Array to store valid moves</param>
void Player::getTransitions(bool m[NUM_WORKERS_PER_PLAYER][8][8]) {
    //initialize as 0
    for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
        for (int move = 0; move < 8; move++) {
            for (int placement = 0; placement < 8; placement++) {
                m[w][move][placement] = 0;
            }
        }
    }

    int playerw[2];
    getWorkers(playerw);
    
    int tempM[8]; //temporary array of valid moves
    int tempP[8]; //temporary array of valid placements

 
    Worker tempWorker;

    //getting moves of worker 0;
    tempWorker = W[playerw[0]];
    tempWorker.getMoves(tempM);

    for (int move = 0; move < 8; move++) {
        if (tempM[move] == 0) { //cannot move this direction
            for (int placement = 0; placement < 8; placement++) {
                m[0][move][placement] = 0;
            }
        }
        else { //can move in this direction
            tempWorker = W[playerw[0]]; //reset to actual position of worker
            tempWorker.makeMove(move); //move it.
            tempWorker.getPlacements(tempP); //populate the valid placements
            for (int placement = 0; placement < 8; placement++) {
                if (move + placement == 7) { //allowing placement "over" last position
                    m[0][move][placement] = 1;
                }
                else
                    m[0][move][placement] = tempP[placement];
            }
        }
    }

    //getting moves of worker 1
    tempWorker = W[playerw[1]];
    tempWorker.getMoves(tempM);

    for (int move = 0; move < 8; move++) {
        if (tempM[move] == 0) { //cannot move this direction
            for (int placement = 0; placement < 8; placement++) {
                m[1][move][placement] = 0;
            }
        }
        else { //can move in this direction
            tempWorker = W[playerw[1]]; //reset to actual position of worker
            tempWorker.makeMove(move); //move it.
            tempWorker.getPlacements(tempP); //populate the valid placements
            for (int placement = 0; placement < 8; placement++) {
                if (move + placement == 7) { //allowing placement "over" last position
                    m[1][move][placement] = 1;
                }
                else
                    m[1][move][placement] = tempP[placement];
            }
        }
    }
}
/// <summary>
/// Prints out literal interpretation of getTransitions().
/// </summary>
void Player::printTransitions() {
    bool m[NUM_WORKERS_PER_PLAYER][8][8];
    getTransitions(m);
    
    for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
        cout << "WORKER " << w << endl;
        for (int move = 0; move < 8; move++) {
            for (int placement = 0; placement < 8; placement++) {
                cout << m[w][move][placement] << " ";
            }
            cout << endl;
        }

    }
    int count_moves = 0;
    int count_placements = 0;

    int workers[NUM_WORKERS_PER_PLAYER];
    getWorkers(workers);


    for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
        cout << "------" << "  WORKER " << workers[w] << "  " << "------" << endl;
        for (int move = 0; move < 8; move++) {
            count_placements = 0;
            for (int placement = 0; placement < 8; placement++) {
                if (m[w][move][placement] == 1) {
                    count_placements++;
                }
            }
            if (count_placements > 0) {
                count_moves++;
                cout << "CAN MOVE " << move << " with placements: " << count_placements << endl;
            }
            else {
                cout << "CANNOT MOVE " << move << endl;
            }
        }
    }
}


//Free functions
/// <summary>
/// Checks provided location for worker occupant.
/// </summary>
/// <param name="x">x coordinate in Board[x][y]</param>
/// <param name="y">y coordinate in Board[x][y]</param>
/// <returns>index of worker if occupied, -1 if unoccupied.</returns>
int checkWorkers(int x, int y) {
    for (int i = 0; i < NUM_WORKERS; i++) {
        if (W[i].x == x && W[i].y == y)
            return i;
    }
    return -1;
}
/// <summary>
/// Checks provided location for worker occupant.
/// </summary>
/// <param name="x">x coordinate in Board[x][y]</param>
/// <param name="y">y coordinate in Board[x][y]</param>
/// <returns>Whether or not if occupied.</returns>
bool checkOccupied(int x, int y) {
    for (int i = 0; i < NUM_WORKERS; i++) {
        if (W[i].x == x && W[i].y == y)
            return true;
    }
    return false;
}
/// <summary>
/// Checks for valid height differential.
/// </summary>
/// <param name="x">current x coordinate in Board[x][y]</param>
/// <param name="y">current y coordinate in Board[x][y]</param>
/// <param name="x2">new x coordinate in Board[x][y]</param>
/// <param name="y2">new y coordinate in Board[x][y]</param>
/// <returns> Whether or not height difference is at most 1 higher.</returns>
bool checkLevels(int x, int y, int x2, int y2) {
    return (Board[x2][y2] <= Board[x][y] + 1);
}
/// <summary>
/// Checks for complete tower (height = 4).
/// </summary>
/// <param name="x">x coordinate in Board[x][y]</param>
/// <param name="y">y coordinate in Board[x][y]</param>
/// <returns>Whether or not height is 4.</returns>
bool checkFourth(int x, int y) {
    return (Board[x][y] == 4);
}
/*///////////////////////////////////////////////////////////////////
exportBoardState()
- exports the boardstate of Board and W arrays into provided stream.
*////////////////////////////////////////////////////////////////////
void exportBoardStateStream(int turn, ostream& out) {
    int index;
    out << turn << " ";
    for (int i = 0; i < BOARD_SIZE; i++) {
        for (int j = 0; j < BOARD_SIZE; j++) {
            out << Board[i][j] << " ";
        }
    }
    for (int i = 0; i < BOARD_SIZE; i++) {
        for (int j = 0; j < BOARD_SIZE; j++) {
            index = checkWorkers(i, j);
            if (index == -1)
                out << 0 << " ";
            else {
                if (W[index].color == turn) {
                    out << 1 << " ";
                }
                else {
                    out << -1 << " ";
                }
            }
        }
    }
    out << endl;
}
/*///////////////////////////////////////////////////////////////////
exportBoardState()
- exports the boardstate of Board and W arrays into provided string
*////////////////////////////////////////////////////////////////////
void exportBoardState(int turn, string& out) {
    int index;
    out = "";
    out.append(to_string(turn)+" ");
    for (int i = 0; i < BOARD_SIZE; i++) {
        for (int j = 0; j < BOARD_SIZE; j++) {
            out.append(to_string(Board[i][j]));
            out.append(" ");
        }
    }
    for (int i = 0; i < BOARD_SIZE; i++) {
        for (int j = 0; j < BOARD_SIZE; j++) {
            index = checkWorkers(i, j);
            if (index == -1) {
                out.append("0 ");
            }
            else {
                if (W[index].color == turn) {
                    //cout << "true " << W[index].color << endl;
                    out.append("1 ");
                }
                else {
                    out.append("-1 ");
                }
            }
        }
    }
    //cout << out << endl;
}
/// <summary>
/// Populates Board and W with provided input boardstate
/// </summary>
/// <param name="turn">Color of player whose point of view</param>
/// <param name="p1">Player 1</param>
/// <param name="p2">Player 2</param>
/// <param name="in">Input stream with boardstate</param>
void importBoardState(int& turn, Player *p1, Player *p2, istream& in) {
    in >> turn;
    if (turn == p1->getColor())
        turn = 0;
    else
        turn = 1;
    for (int i = 0; i < BOARD_SIZE; i++) {
        for (int j = 0; j < BOARD_SIZE; j++) {
            in >> Board[i][j];
        }
    }

    int temp;
    int index = 0;
    for (int i = 0; i < BOARD_SIZE; i++) {
        for (int j = 0; j < BOARD_SIZE; j++) {
            in >> temp;
            if (temp == 1) {
                W[index].x = i;
                W[index].y = j;

                if (turn == 0) {
                    W[index].color = p1->getColor();
                }
                else {
                    W[index].color = p2->getColor();
                }
                index++;
            }
            if (temp == -1) {
                W[index].x = i;
                W[index].y = j;

                if (turn == 0) {
                    W[index].color = p2->getColor();
                }
                else {
                    W[index].color = p1->getColor();
                }
                index++;
            }
        }
    }
}
/// <summary>
/// Populates Board and W with provided vector boardstate
/// </summary>
/// <param name="turn">Color of player whose point of view.</param>
/// <param name="p1">Player 1</param>
/// <param name="p2">Player 2</param>
/// <param name="v">Boardstate</param>
void importBoardState(int& turn, Player* p1, Player* p2, double v[50]) {
    //converts turn from color of player to 0 or 1

    if (turn == p1->getColor())
        turn = 0;
    else
        turn = 1;

    for (int i = 0; i < 25; i++) {
        Board[i / 5][i % 5] = v[i];
    }


    int index = 0;
    for (int i = 25; i < 50; i++) {
        //cout << index << endl;
        if (v[i] == 1) {
            W[index].x = (i - 25) / 5;
            W[index].y = (i - 25) % 5;
            if(turn == 0)
                W[index].color = p1->getColor();
            else
                W[index].color = p2->getColor();
            index++;
        }
        else if (v[i] == -1) {
            W[index].x = (i - 25) / 5;
            W[index].y = (i - 25) % 5;
            if (turn == 0)
                W[index].color = p2->getColor();
            else
                W[index].color = p1->getColor();
            index++;
        }
        if (index == 4)
            break;
    }

  /*  cout << "test" << endl;
    if (turn == p1->getColor())
        turn = 0;
    else
        turn = 1;

    for (int i = 0; i < 25; i++) {
        int x = i % 5;
        int y = i / 5;
        Board[x][y] = v[i];
    }
    int index = 0;
    for (int i = 25; i < 50; i++) {
        int x = (i-25) % 5;
        int y = (i-25) / 5;
        if (v[i] == 1) {
            W[index].x = x;
            W[index].y = y;
            W[index].color = p1->getColor();
        }
        else if(v[i] == -1) {
            W[index].x = x;
            W[index].y = y;
            W[index].color = p2->getColor();
        }

    }

    cout << "final" << endl;*/

}
/*///////////////////////////////////////////////////////////////////
extractBoardState()
- populates an vector with the boardstate of Board and W.
*////////////////////////////////////////////////////////////////////
void extractBoardState(int turn, double v[50]) {
    int index = 0;
    int temp;

    for (int i = 0; i < BOARD_SIZE; i++) {
        for (int j = 0; j < BOARD_SIZE; j++) {
            v[index] = Board[i][j];
            index++;
        }
    }
    for (int i = 0; i < BOARD_SIZE; i++) {
        for (int j = 0; j < BOARD_SIZE; j++) {
            temp = checkWorkers(i, j);
            if (temp == -1) {
                v[index] = 0;
            }
            else {
                if (W[temp].color == turn) {
                    v[index] = 1;
                }
                else {
                    v[index] = -1;
                }
            }
            index++;
        }
    }

}
/*///////////////////////////////////////////////////////////////////
extractBoardState()
- populates an vector with the boardstate after move and placement. 
*////////////////////////////////////////////////////////////////////
void extractNewBoardState(int worker, int move, int placement, int turn, double v[50]) {
    int index = 0;
    int temp;
    
    int X = W[worker].x;
    int Y = W[worker].y;

    int moveX = X;
    int moveY = Y;

    //get new move positions
    adjacentSpace(move, moveX, moveY);

    int placeX = moveX;
    int placeY = moveY;

    //get new placement positions
    adjacentSpace(placement, placeX, placeY);
    
    for (int i = 0; i < BOARD_SIZE; i++) {
        for (int j = 0; j < BOARD_SIZE; j++) {
            v[index] = Board[i][j];
            if (i == placeX && j == placeY) {
                v[index]++;
            }
            index++;
        }
    }
    for (int i = 0; i < BOARD_SIZE; i++) {
        for (int j = 0; j < BOARD_SIZE; j++) {
            temp = checkWorkers(i, j);
            if ((temp == -1 || temp == worker) && !(i == moveX && j == moveY)) {
                v[index] = 0;
            }
            else {
                if (i == moveX && j == moveY)
                    temp = worker;
                if (W[temp].color == turn) {
                    v[index] = -1;
                }
                else {
                    v[index] = 1;
                }
            }
            index++;
        }
    }

}
/*///////////////////////////////////////////////////////////////////
printVector()
- prints a vector (for testing ANN)
*////////////////////////////////////////////////////////////////////
void printVector(double v[50]) {
    for (int i = 0; i < 50; i++) {
        cout << v[i] << " ";
    }
    cout << endl;
}
/*///////////////////////////////////////////////////////////////////
adjacentSpace()
- encodes a shift of (x,y) to an adjacent space as a move.
*////////////////////////////////////////////////////////////////////
bool adjacentSpace(string move, int& x, int& y) {
	//check that initial coordinates are within bounds
	if (x < 0 || x >= BOARD_SIZE || y < 0 || y >= BOARD_SIZE) {
		cout << "Coordinates provided are out of bounds" << endl;
		return false;
	}
	
    if (move == "UL" && x > 0 && y > 0) { //upperLeft
        x -= 1;
        y -= 1;
    }
    else if (move == "UR" && x > 0 && y < 4) { //upperRight
        x -= 1;
        y += 1;
    }
    else if (move == "LL" && x < 4 && y > 0) { //lowerLeft
        x += 1;
        y -= 1;
    }
    else if (move == "LR" && x < 4 && y < 4) { //lowerRight
        x += 1;
        y += 1;
    }
    else if ((move == "UM" || move == "up") && x > 0) { //upperMiddle / up
        x -= 1;
    }
    else if ((move == "LM" || move == "down") && x < 4) { //lowerMiddle / down
        x += 1;
    }
    else if ((move == "CL" || move == "left") && y > 0) { //centerLeft / left
        y -= 1;
    }
    else if ((move == "CR" || move == "right") && y < 4) { //centerRight / right
        y += 1;
    }
    else {
        return false;
    }
    return true;
}
/*///////////////////////////////////////////////////////////////////
adjacentSpace()
- encodes a shift of (x,y) to an adjacent space as a move [0,7].
*////////////////////////////////////////////////////////////////////
bool adjacentSpace(int move, int& x, int& y) {
    //check that initial coordinates are within bounds
    if (x < 0 || x >= BOARD_SIZE || y < 0 || y >= BOARD_SIZE) {
        cout << "Coordinates provided are out of bounds" << endl;
        return false;
    }

    if (move == 0 && x > 0 && y > 0) { //upperLeft
        x -= 1;
        y -= 1;
    }
    else if (move == 1 && x > 0) { //upperMiddle / up
        x -= 1;
    }
    else if (move == 2 && x > 0 && y < 4) { //upperRight
        x -= 1;
        y += 1;
    }
    else if (move == 3 && y > 0) { //centerLeft / left
        y -= 1;
    }
    else if (move == 4 && y < 4) { //centerRight / right
        y += 1;
    }
    else if (move == 5 && x < 4 && y > 0) { //lowerLeft
        x += 1;
        y -= 1;
    }
    else if (move == 6 && x < 4) { //lowerMiddle / down
        x += 1;
    }
    else if (move == 7 && x < 4 && y < 4) { //lowerRight
        x += 1;
        y += 1;
    }
    else {
        return false;
    }
    return true;
}
/*///////////////////////////////////////////////////////////////////
printBoard()
- Prints Board array
*////////////////////////////////////////////////////////////////////
void printBoard(int space, ostream& out) {
    char ch = 'A'; //iterates {A, B, ... , N} for top border
    int num = 1; //iterates {1, 2, ... , n} for side border
    int w; //index for Workers[]

    //printing top border {***}
    setColor(BORDER_COLOR);
    for (int i = 0; i < BOARD_SIZE + 2; i++) {
        if (i == 0 || i == BOARD_SIZE + 1) //prints '*' on first and last of top border
            out << setw(space) << "*";
        else { //prints {A, B, ... , N} in center of top border
            out << setw(space) << ch;
            ch++;
        }
    }
    out << endl;

    //printing the board
    for (int i = 0; i < BOARD_SIZE; i++) {
        out << setw(space) << num;
        num++;
        setColor(BOARD_COLOR);
        for (int j = 0; j < BOARD_SIZE; j++) {
            w = checkWorkers(i, j);
            if (w != -1) {
                setColor(W[w].color);
            }
            out << setw(space) << Board[i][j];
            //out << setw(space) << i << j;
            if (w != -1) {
                setColor(BOARD_COLOR);
            }
        }
        setColor(BORDER_COLOR);
        out << setw(space) << "*";
        out << endl;
    }

    //printing bottom border {***}
    for (int i = 0; i < BOARD_SIZE + 2; i++) {
        out << setw(space) << "*";
    }
    out << endl;

    setColor(DEFAULT_COLOR);

    return;
}
/*///////////////////////////////////////////////////////////////////
setColor()
- changes cout text color.
*////////////////////////////////////////////////////////////////////
void setColor(int color) {
    HANDLE hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
    switch (color) {
    case WHITE:
        SetConsoleTextAttribute(hConsole, FOREGROUND_INTENSITY | FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE); //WHITE
        break;
    case RED:
        SetConsoleTextAttribute(hConsole, FOREGROUND_INTENSITY | FOREGROUND_RED); //RED
        break;
    case DARKRED:
        SetConsoleTextAttribute(hConsole, FOREGROUND_RED); //DARKRED
        break;
    case GREEN:
        SetConsoleTextAttribute(hConsole, FOREGROUND_INTENSITY | FOREGROUND_GREEN); //GREEN
        break;
    case DARKGREEN:
        SetConsoleTextAttribute(hConsole, FOREGROUND_GREEN); //DARKGREEN
        break;
    case BLUE:
        SetConsoleTextAttribute(hConsole, FOREGROUND_INTENSITY | FOREGROUND_BLUE); //BLUE
        break;
    case DARKBLUE:
        SetConsoleTextAttribute(hConsole, FOREGROUND_BLUE); //DARKBLUE
        break;
    case YELLOW:
        SetConsoleTextAttribute(hConsole, FOREGROUND_INTENSITY | FOREGROUND_RED | FOREGROUND_GREEN); //YELLOW
        break;
    case DARKYELLOW:
        SetConsoleTextAttribute(hConsole, FOREGROUND_RED | FOREGROUND_GREEN); //DARKYELLOW
        break;
    case MAGENTA:
        SetConsoleTextAttribute(hConsole, FOREGROUND_INTENSITY | FOREGROUND_RED | FOREGROUND_BLUE); //MAGENTA
        break;
    case DARKMAGENTA:
        SetConsoleTextAttribute(hConsole, FOREGROUND_RED | FOREGROUND_BLUE); //DARKMAGENTA
        break;
    case CYAN:
        SetConsoleTextAttribute(hConsole, FOREGROUND_INTENSITY | FOREGROUND_GREEN | FOREGROUND_BLUE); //CYAN
        break;
    case DARKCYAN:
        SetConsoleTextAttribute(hConsole, FOREGROUND_GREEN | FOREGROUND_BLUE); //DARKCYAN
        break;
    case GRAY:
        SetConsoleTextAttribute(hConsole, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE); //GRAY
        break;
    case DARKGRAY:
        SetConsoleTextAttribute(hConsole, FOREGROUND_INTENSITY); //DARKGRAY
        break;
    case BLACK:
        SetConsoleTextAttribute(hConsole, 0); //BLACK
        break;
    default:
        SetConsoleTextAttribute(hConsole, FOREGROUND_INTENSITY | FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE); //WHITE
    }
    return;
}
/*///////////////////////////////////////////////////////////////////
selectPos()
- pings user to select coordinates on the board using algebraic 
notation.
*////////////////////////////////////////////////////////////////////
void selectPos(int& x, int& y, string ask) {
    char ch; //column index input
    int num; //row index input
    if (ask != "") //print prompt if given. 
        cout << ask << endl;
    while (true) {
        cin >> ch;
        if (cin.fail()) { //checks to see if ch is not a char
            cout << "Invalid choice." << endl;
            cin.clear();
            cin.ignore(INT64_MAX, '\n');
        }
        else {
            ch = tolower(ch); //converts ch to its lowercase counterpart
            if (!(ch - 'a' < BOARD_SIZE && ch - 'a' >= 0)) { //ensures that selection is within range
                cout << "Invalid choice." << endl;
            }
            else {
                y = ch - 'a'; //assigns y the ordinal value of input ch.
                cin >> num;
                if (cin.fail() || num < 1 || num > BOARD_SIZE) { //checks to see if num is not int or not in range
                    cout << "Invalid choice." << endl;
                    cin.clear();
                    cin.ignore(INT64_MAX, '\n');
                }
                else {
                    x = num - 1; //assigns x the ordinal value of input num. 
                    break;
                }
            }
        }
    }
    return;
}

//Worker functions
/// <summary>
/// Populates m with valid moves from current worker position.
/// </summary>
/// <param name="m">Array referencing adjacent positions.</param>
/// <returns>If at least one viable move</returns>
bool Worker::getMoves(int m[8]) {
    int posX = x;
    int posY = y;
    bool move;
    bool isMove = false;

    for (int i = 0; i < 8; i++) {
        m[i] = 0;
        posX = x;
        posY = y;
        move = adjacentSpace(i, posX, posY);
        if (move) { //if there is an adjacent space
            if (!checkOccupied(posX, posY)) { //if the space is unoccupied
                if (checkLevels(x,y,posX,posY) && !checkFourth(posX,posY)) { //if position at most 1 level higher
                    m[i] = 1; //mark as a valid move
                    isMove = true;
                }
            }
        }
    }
    return isMove;
}
/// <summary>
/// Populates m with valid placements from current worker position.
/// </summary>
/// <param name="m">Array referencing adjacent positions.</param>
/// <returns>If at least one viable placement</returns>
bool Worker::getPlacements(int m[8]) {
    int posX = x;
    int posY = y;
    bool move;
    bool isPlacement = false;

    for (int i = 0; i < 8; i++) {
        m[i] = 0;
        posX = x;
        posY = y;

        move = adjacentSpace(i, posX, posY);
        if (move) { //if there is an adjacent space
            if (!checkOccupied(posX, posY)) { //if the space is unoccupied
                if (!checkFourth(posX, posY)) { //if position is not complete tower
                    m[i] = 1; //mark as a valid move
                    isPlacement = true;
                }
            }
        }
    }
    return isPlacement;
}
/// <summary>
/// Updates worker position using adjacent move.
/// </summary>
/// <param name="move">Encoded adjacent tile from relative position</param>
void Worker::makeMove(int move) {
    adjacentSpace(move, x, y);
    return;
}
/// <summary>
/// Updates Board height using adjacent placement.
/// </summary>
/// <param name="move">Encoded adjacent tile from relative position</param>
void Worker::makePlacement(int move) {
    int px = x;
    int py = y;
    adjacentSpace(move, px, py);
    Board[px][py]++;
    return;
}
/// <summary>
/// (Unused) Updates worker position using adjacent move.
/// </summary>
/// <param name="cx">Change in x</param>
/// <param name="cy">Change in y</param>
void Worker::makeMove(int cx,int cy) {
    x = cx;
    y = cy;
    return;
}
/// <summary>
/// Checks to see if there is a valid move.
/// </summary>
/// <returns>Whether or not there is at least one move</returns>
bool Worker::isMove() {
    int level = Board[x][y]; //current level of the worker
    int tempX, tempY;

    for (int i = 0; i < 3; i++) { //iterates Top, Center, Bottom
        for (int j = 0; j < 3; j++) { //iterates Left, Middle, Right
            tempX = x - 1 + i;
            tempY = y - 1 + j;

            if (tempX < BOARD_SIZE && tempX >= 0 && tempY < BOARD_SIZE && tempY >= 0) { //checks to see within bounds
                if (checkWorkers(tempX, tempY) == -1) { //checks to see if space is unoccupied
                    if (Board[tempX][tempY] <= level + 1 && Board[tempX][tempY] != 4) { //checks to see if level is at most one higher
                        return true;
                    }
                }
            }
        }
    }
    return false;
}

//plyrHuman functions
/// <summary>
/// Takes a turn using player input.
/// </summary>
/// <returns>1 if win, -1 if lose, 0 if neutral</returns>
int plyrHuman::takeTurn() {
    int x, y;
    int worker;
    int index;
    int w[NUM_WORKERS_PER_PLAYER];

    getWorkers(w); //populates w with indexes of player's workers

    if (!W[w[0]].isMove() && !W[w[1]].isMove()) {
        return -1;
    }

    //selecting a worker.
    bool check = false;
    while (!check) {
        selectPos(x, y, getName() + ", select a worker.");
        index = checkWorkers(x, y);
        if (index == -1) {
            cout << "Must select an occupied space..." << endl;
        }
        else if (W[index].color != getColor()) {
            cout << "Must select one of your workers..." << endl;
        }
        else if (!W[index].isMove()) {
            cout << "Must select a worker that can move..." << endl;
        }
        else {
            check = true;
        }
    }

    //selecting a tile to move on.
    check = false;
    while (!check) {
        selectPos(x, y, "Move to which tile?");
        worker = checkWorkers(x, y);
        if (worker != -1) { //checks to see if space is occupied.
            cout << "Cannot move onto an occupied space..." << endl;
        }
        else if (abs(x - W[index].x) > 1 || abs(y - W[index].y) > 1) { //checks to see if adjacent
            cout << "Must be within 1 square of worker..." << endl;
        }
        else if (Board[x][y] - Board[W[index].x][W[index].y] > 1) {
            cout << "Must be at most 1 level upwards..." << endl;
        }
        else if (Board[x][y] == 4) {
            cout << "Cannot move onto a dome..." << endl;
        }
        else {
            check = true;
        }
    }
    //moving the worker
    W[index].makeMove(x, y);

    //checks if player won the game.
    if (Board[W[index].x][W[index].y] == 3) {
        return 1;
    }

    //selecting a tile to build on.
    check = false;
    while (!check) {
        selectPos(x, y, "Build on which tile?");
        worker = checkWorkers(x, y); //check to see if space is occupied

        if (worker != -1) { //checks to see if space is occupied.
            cout << "Cannot build on an occupied space..." << endl;
        }
        else if ((abs(x - W[index].x) > 1 || abs(y - W[index].y) > 1)) {
            cout << "Must be within 1 square of worker..." << endl;
        }
        else if (Board[x][y] == 4) {
            cout << "Cannot build on a complete tower..." << endl;
        }
        else {
            Board[x][y]++;
            check = true;
        }
    }
    return 0;
}
/// <summary>
/// Place a worker on Board using player input.
/// </summary>
/// <param name="index">Index of worker in W.</param>
void plyrHuman::setWorker(int index) {
    int x, y;
    bool check = true;
    while (check) {
        check = false;
        selectPos(x, y, getName() + ", select a position for your worker.");
        for (int i = 0; i < NUM_WORKERS; i++) {
            if (W[i].x == x && W[i].y == y) {
                cout << "Invalid choice." << endl;
                check = true;
                break;
            }
        }
    }
    W[index].makeMove(x, y);
    W[index].color = getColor();
}

//plyrRandom functions
/// <summary>
/// Takes a turn using random selection.
/// </summary>
/// <returns>1 if win, -1 if lose, 0 if neutral</returns>
int plyrRandom::takeTurn() {
    int w[NUM_WORKERS_PER_PLAYER];
    getWorkers(w); //populates w with indexes of player's workers

    int index;
    //cout << w[0] << " " << w[1] << endl;

    int w0moves[8];
    bool canw0 = W[w[0]].getMoves(w0moves);
    int w1moves[8];
    bool canw1 = W[w[1]].getMoves(w1moves);
    if (!canw0 && !canw1) { //neither have a valid move
        return -1;
    }
    else if (canw0 && canw1) { //randomly select one if both have a valid move.
        index = rand() % 2;
    }
    else if (canw0) { //select to move worker 0
        index = 0;
    }
    else { //select to move worker 1
        index = 1;
    }

    //cout << "past if/else" << endl;
    //cout << canw0 << " " << canw1 << endl;
    
    int move = rand() % 8;
    int x, y;
    if (index == 0) {
        //cout << "index 0" << endl;
        while (w0moves[move] == 0) {
            move = rand() % 8;
        }
        adjacentSpace(move, W[w[0]].x, W[w[0]].y);
        W[w[0]].getPlacements(w0moves);
        move = rand() % 8;
        while (w0moves[move] == 0) {
            move = rand() % 8;
        }
        x = W[w[0]].x;
        y = W[w[0]].y;
        if (Board[x][y] == 3) {
            return 1;
        }

        adjacentSpace(move, x, y);
        Board[x][y]++;
    }
    else {
        //cout << "index 1" << endl;
        while (w1moves[move] == 0) {
            move = rand() % 8;
        }
        adjacentSpace(move, W[w[1]].x, W[w[1]].y);
        W[w[1]].getPlacements(w1moves);
        move = rand() % 8;
        while (w1moves[move] == 0) {
            move = rand() % 8;
        }
        x = W[w[1]].x;
        y = W[w[1]].y;
        if (Board[x][y] == 3) {
            return 1;
        }
        adjacentSpace(move, x, y);
        Board[x][y]++;
    }

    return 0;
}
/// <summary>
/// Place a worker on Board using random selection.
/// </summary>
/// <param name="index">Index of worker in W.</param>
void plyrRandom::setWorker(int index) {
    int x, y;
    bool check = true;
    while (check) { //gets random coordinates until unoccupied space is found.
        check = false;
        x = rand() % BOARD_SIZE;
        y = rand() % BOARD_SIZE;
        for (int i = 0; i < NUM_WORKERS; i++) {
            if (W[i].x == x && W[i].y == y) {
                check = true;
                break;
            }
        }
    }
    W[index].x = x;
    W[index].y = y;
    W[index].color = getColor();
}

//plyrHeuristic functions
/// <summary>
/// Printing all possible moves.
/// </summary>
int plyrHeuristic::printTurns() {
    /*int w[NUM_WORKERS_PER_PLAYER];
    getWorkers(w); //populates w with indexes of player's workers

    int index;
    cout << w[0] << " " << w[1] << endl;

    int w0moves[8];
    bool canw0 = W[w[0]].getMoves(w0moves);
    int w1moves[8];
    bool canw1 = W[w[1]].getMoves(w1moves);

    if (!canw0 && !canw1) { //neither have a valid move
        return -1;
    }
    
    int x, y;
    index = -1;
    for (int i = 0; i < 8; i++) {
        if (w0moves[i] == 1) {
            x = W[w[0]].x;
            y = W[w[0]].y;
            adjacentSpace(i, x, y);
            if (Board[x][y] > Board[W[w[0]].x][W[w[0]].y]) {
                adjacentSpace(i, W[w[0]].x, W[w[0]].y);
                index = 0;
                break;
            }
        }
        else if (w1moves[i] == 1) {
            x = W[w[1]].x;
            y = W[w[1]].y;
            adjacentSpace(i, x, y);
            if (Board[x][y] > Board[W[w[1]].x][W[w[1]].y]) {
                adjacentSpace(i, W[w[1]].x, W[w[1]].y);
                index = 1;
                break;
            }
        }
    }
    int move = rand() % 8;
    if (index == -1) {
        if (canw0) {
            while (w0moves[move] == 0) {
                move = rand() % 8;
            }
            adjacentSpace(move, W[w[0]].x, W[w[0]].y);
            index = 0;
        }
        else {
            int move = rand() % 8;
            while (w1moves[move] == 0) {
                move = rand() % 8;
            }
            adjacentSpace(move, W[w[1]].x, W[w[1]].y);
            index = 1;
        }
    }

    if (index == 0) {
        W[w[0]].getPlacements(w0moves);
        move = rand() % 8;
        while (w0moves[move] == 0) {
            move = rand() % 8;
        }
        x = W[w[0]].x;
        y = W[w[0]].y;
        if (Board[x][y] == 3) {
            return 1;
        }

        adjacentSpace(move, x, y);
        Board[x][y]++;
    }
    else {
        W[w[1]].getPlacements(w1moves);
        move = rand() % 8;
        while (w1moves[move] == 0) {
            move = rand() % 8;
        }
        x = W[w[1]].x;
        y = W[w[1]].y;
        if (Board[x][y] == 3) {
            return 1;
        }

        adjacentSpace(move, x, y);
        Board[x][y]++;
    }

    return 0;
    */

printTransitions();

//getWorkers for workers
int workers[NUM_WORKERS_PER_PLAYER];
getWorkers(workers);
//getTransitions for bool array
bool moves[NUM_WORKERS_PER_PLAYER][8][8];
getTransitions(moves);
//extractNewBoardState for respective boardstates
double* boardstates[NUM_WORKERS_PER_PLAYER][8][8];
for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
    for (int m = 0; m < 8; m++) {
        for (int p = 0; p < 8; p++) {
            if (moves[w][m][p] == 1) { //there is a move
                double* v = new double[50];
                extractNewBoardState(workers[w], m, p, getColor(), v);
                boardstates[w][m][p] = v;

            }
        }
    }
}

//printing to verify correctness
for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
    for (int m = 0; m < 8; m++) {
        for (int p = 0; p < 8; p++) {
            if (moves[w][m][p] == 0) {
                cout << w << ":" << "[" << m << "," << p << "]";
                cout << "N/A" << endl;
            }
            else {
                cout << w << ":" << "[" << m << "," << p << "] ";
                for (int i = 0; i < 50; i++) {
                    cout << boardstates[w][m][p][i] << " ";
                }
                cout << endl;
            }

        }
    }
}

//deallocate memory
for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
    for (int m = 0; m < 8; m++) {
        for (int p = 0; p < 8; p++) {
            if (boardstates[w][m][p] != nullptr) {
                delete[] boardstates[w][m][p];
                boardstates[w][m][p] = nullptr;
            }
        }
    }
}

return 0;
}
/// <summary>
/// Takes a turn using heuristic evaluation.
/// </summary>
/// <returns></returns>
int plyrHeuristic::takeTurn() {

    //getWorkers for workers
    int workers[NUM_WORKERS_PER_PLAYER];
    getWorkers(workers);
    if (!W[workers[0]].isMove() && !W[workers[1]].isMove())
        return -1;

    //getTransitions for bool array
    bool moves[NUM_WORKERS_PER_PLAYER][8][8];
    getTransitions(moves);
    //extractNewBoardState for respective boardstates
    double* boardstates[NUM_WORKERS_PER_PLAYER][8][8];
    for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
        for (int m = 0; m < 8; m++) {
            for (int p = 0; p < 8; p++) {
                if (moves[w][m][p] == 1) { //there is a move
                    double* v = new double[50];
                    extractNewBoardState(workers[w], m, p, getColor(), v);
                    boardstates[w][m][p] = v;
                }
                else {
                    boardstates[w][m][p] = nullptr;
                }

            }
        }
    }

  
    //new array to fill up with board evaluations
    double eval[NUM_WORKERS_PER_PLAYER][8][8];
    for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
        for (int m = 0; m < 8; m++) {
            for (int p = 0; p < 8; p++) {
                eval[w][m][p] = EvalFcn(boardstates[w][m][p]);
                //cout << w << " " << m << " " << p << ": " << setprecision(4) << eval[w][m][p] << endl;
                //cout << "[" << w << m << p << "]: " << eval[w][m][p] << " == " << EvalFcn(boardstates[w][m][p]) << endl;
            }
        }
    }


    double max = 0;
    //find max
    for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
        for (int m = 0; m < 8; m++) {
            for (int p = 0; p < 8; p++) {
                if (eval[w][m][p] > max)
                    max = eval[w][m][p];
            }
        }
    }

 
    threeIndex turns[128];
    int num_turns = 0;

    //populate array with moves within threshold of max
    for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
        for (int m = 0; m < 8; m++) {
            for (int p = 0; p < 8; p++) {
                if (eval[w][m][p] >= max - threshold && moves[w][m][p] == 1) {
                    turns[num_turns].x = w;
                    turns[num_turns].y = m;
                    turns[num_turns].z = p;
                    num_turns++;
                }
            }
        }
    }


    if (num_turns == 0) {
        int x = rand()%NUM_WORKERS_PER_PLAYER; int y=rand() %8; int z = rand()% 8;
        while (moves[x][y][z] == 0) {
            x = rand() % NUM_WORKERS_PER_PLAYER;
            y = rand() % 8; 
            z = rand() % 8;
        }
        turns[num_turns].x = x;
        turns[num_turns].y = y;
        turns[num_turns].z = z;
        num_turns++;

        //cout << x << y << z << endl;
    }
   // cout << "TEST " << num_turns << " " << max << endl;

    int make_turn = rand() % num_turns;
    W[workers[turns[make_turn].x]].makeMove(turns[make_turn].y);
    W[workers[turns[make_turn].x]].makePlacement(turns[make_turn].z);

    int win = 0;

    if (Board[W[workers[turns[make_turn].x]].x][W[workers[turns[make_turn].x]].y] == 3) {
        win = 1;
    }

    // Deallocate memory
    for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
        for (int m = 0; m < 8; m++) {
            for (int p = 0; p < 8; p++) {
                if (boardstates[w][m][p] != nullptr) {
                    delete[] boardstates[w][m][p];
                    boardstates[w][m][p] = nullptr;  
                }
            }
        }
    }

    return win;

}
/// <summary>
/// Evaluates a board position (UPDATES V).
/// </summary>
/// <param name="v">Boardstate</param>
/// <returns>Value estimation</returns>
double plyrHeuristic::EvalFcn(double v[50]) {

    if (v == nullptr)
        return 0;

    int index = 0; 

    int workers[4];
    for (int i = 25; i < 50; i++) {
        if (v[i] == 1) {
            workers[index] = i;
            index++;
        }
    }
    for (int i = 25; i < 50; i++) {
        if (v[i] == -1) {
            workers[index] = i;
            index++;
        }
    }

    double height1 = v[workers[0] - 25] + v[workers[1] - 25];
    double height2 = v[workers[2] - 25] + v[workers[3] - 25];
    double height_difference = height1 - height2;

    double win_lose = 0;
    if (v[workers[0] - 25] == 3 || v[workers[1] - 25] == 3) {
        win_lose = 50;
    }
    else if (v[workers[2] - 25] == 3 || v[workers[3] - 25] == 3) {
        win_lose = -50;
    }

    double centricity = 0;
    if (workers[0] - 25 == 12) {
        centricity += 3;
    }
    else if (workers[0] - 25 > 4 && workers[0] - 25 < 20 && (workers[0] - 25) % 5 != 0 && (workers[0] - 25) % 5 != 4) {
        centricity += 1;
    }
    if (workers[1] - 25 == 12) {
        centricity += 3;
    }
    else if (workers[1] - 25 > 4 && workers[1] - 25 < 20 && (workers[1] - 25) % 5 != 0 && (workers[1] - 25) % 5 != 4) {
        centricity += 1;
    }

    return (height_difference * 10) + centricity + win_lose + 2.5;
}
/// <summary>
/// Evaluates a Boardstate.
/// </summary>
/// <param name="x">const Boardstate</param>
/// <returns>Value estimation</returns>
double plyrHeuristic::getEval(double x[50]) {
    double v[50];
    for (int i = 0; i < 50; i++) {
        v[i] = x[i];
    }
    
    if (v == nullptr)
        return 0;

    int index = 0;

    int workers[4];
    for (int i = 25; i < 50; i++) {
        if (v[i] == 1) {
            workers[index] = i;
            index++;
        }
    }
    for (int i = 25; i < 50; i++) {
        if (v[i] == -1) {
            workers[index] = i;
            index++;
        }
    }

    double height1 = v[workers[0] - 25] + v[workers[1] - 25];
    double height2 = v[workers[2] - 25] + v[workers[3] - 25];
    double height_difference = height1 - height2;

    double win_lose = 0;
    if (v[workers[0] - 25] == 3 || v[workers[1] - 25] == 3) {
        win_lose = 50;
    }
    else if (v[workers[2] - 25] == 3 || v[workers[3] - 25] == 3) {
        win_lose = -50;
    }

    double centricity = 0;
    if (workers[0] - 25 == 12) {
        centricity += 3;
    }
    else if (workers[0] - 25 > 4 && workers[0] - 25 < 20 && (workers[0] - 25) % 5 != 0 && (workers[0] - 25) % 5 != 4) {
        centricity += 1;
    }
    if (workers[1] - 25 == 12) {
        centricity += 3;
    }
    else if (workers[1] - 25 > 4 && workers[1] - 25 < 20 && (workers[1] - 25) % 5 != 0 && (workers[1] - 25) % 5 != 4) {
        centricity += 1;
    }

    return (height_difference * 10) + centricity + win_lose + 2.5;
}
/// <summary>
/// Place a worker on Board using random selection.
/// </summary>
/// <param name="index">Index of worker in W.</param>
void plyrHeuristic::setWorker(int index) {
    int x, y;
    bool check = true;
    while (check) { //gets random coordinates until unoccupied space is found.
        check = false;
        x = rand() % BOARD_SIZE;
        y = rand() % BOARD_SIZE;
        for (int i = 0; i < NUM_WORKERS; i++) {
            if (W[i].x == x && W[i].y == y) {
                check = true;
                break;
            }
        }
    }
    W[index].x = x;
    W[index].y = y;
    W[index].color = getColor();
}

//ANN functions
/// <summary>
/// Player constructor for ANN.
/// </summary>
/// <param name="filename">Input file for ANN.</param>
/// <param name="nm">Name</param>
/// <param name="col">Color</param>
plyrANN::plyrANN(string filename, string nm, int col) {
    name = nm;
    color = col;
    
    for (int i = 0; i < 50; i++) {
        for (int j = 0; j < 50; j++) {
            layer1[i][j] = 0;
            layer2[i][j] = 0;
            layer3[i][j] = 0;
        }
        shift1[i] = 0;
        shift2[i] = 0;
        shift3[i] = 0;
        layer4[0][i] = 0;
    }
   
    ifstream fin;
    fin.open(filename);
    if (!fin.is_open()) {
        cerr << "plyrANN(): Could not open file " << filename << endl;
        exit(EXIT_FAILURE);
    }

    for (int i = 0; i < 50; i++) {
        for (int j = 0; j < 50; j++) {
            fin >> layer1[i][j];
        }
    }
    for (int i = 0; i < 50; i++) {
        fin >> shift1[i];
    }
    for (int i = 0; i < 50; i++) {
        for (int j = 0; j < 50; j++) {
            fin >> layer2[i][j];
        }
    }
    for (int i = 0; i < 50; i++) {
        fin >> shift2[i];
    }
    for (int i = 0; i < 50; i++) {
        for (int j = 0; j < 50; j++) {
            fin >> layer3[i][j];
        }
    }
    for (int i = 0; i < 50; i++) {
        fin >> shift3[i];
    }
    for (int j = 0; j < 50; j++) {
        fin >> layer4[0][j];
    }
}
/// <summary>
/// Outputs ANN internal values.
/// </summary>
/// <param name="out">Output stream</param>
/// <param name="portion">Refers to segment to output</param>
void plyrANN::print(ostream& out, string portion) {
    //layer 1
    if (portion == "ALL" || portion == "L1") {
        for (int i = 0; i < 50; i++) {
            for (int j = 0; j < 50; j++) {
                out << layer1[i][j];
            }
            out << endl;
        }
    }
   
    //shift 1
    if (portion == "ALL" || portion == "S1") {
        for (int i = 0; i < 50; i++) {
            out << shift1[i];
        }
        out << endl;
    }
    
   
    //layer 2
    if (portion == "ALL" || portion == "L2") {
        for (int i = 0; i < 50; i++) {
            for (int j = 0; j < 50; j++) {
                out << layer2[i][j];
            }
            out << endl;
        }
    }
    
    //shift 2
    if (portion == "ALL" || portion == "S2") {
        for (int i = 0; i < 50; i++) {
            out << shift2[i];
        }
        out << endl;
    }

    //layer 3
    if (portion == "ALL" || portion == "L3") {
        for (int i = 0; i < 50; i++) {
            for (int j = 0; j < 50; j++) {
                out << layer3[i][j];
            }
            out << endl;
        }
    }

    //shift 3
    if (portion == "ALL" || portion == "S3") {
        for (int i = 0; i < 50; i++) {
            cout << shift3[i];
        }
        out << endl;
    }

    //layer 4
    if (portion == "ALL" || portion == "L4") {
        for (int i = 0; i < 50; i++) {
            out << layer4[0][i];
        }
        out << endl;
    }
}
/// <summary>
/// Multiplies 50x50 matrix and 50x1 vector
/// </summary>
/// <param name="v">Boardstate or hidden layer</param>
/// <param name="layer">Layer of matrix</param>
void plyrANN::matrixMultiply(double v[50], int layer) {
    double tempv[50];
    double dot = 0;

    for (int i = 0; i < 50; i++) {
        dot = 0;
        for (int j = 0; j < 50; j++) {
            if (layer == 1) {
                dot += layer1[i][j] * v[j];
            }
            else if (layer == 2) {
                dot += layer2[i][j] * v[j];
            }
            else if (layer == 3) {
                dot += layer3[i][j] * v[j];
            }
        }
        tempv[i] = dot;
    }

    for (int i = 0; i < 50; i++) {
        v[i] = tempv[i];
    }

}
/// <summary>
/// Multiplies 1x50 'matrix' with 50x1 vector.
/// </summary>
/// <param name="v">Hidden Layer</param>
/// <returns>Value estimate</returns>
double plyrANN::lastLayerMultiply(double v[50]) {
    double dot = 0;

    for (int i = 0; i < 50; i++) {
        dot += v[i] * layer4[0][i];
    }

    return dot;
}
/// <summary>
/// Shifts vector by a bias
/// </summary>
/// <param name="v">Hidden layer vector</param>
/// <param name="layer">Layer of bias</param>
void plyrANN::shift(double v[50], int layer) {
    for (int i = 0; i < 50; i++) {
        if (layer == 1) {
            v[i] += shift1[i];
        }
        else if (layer == 2) {
            v[i] += shift2[i];
        }
        else if (layer == 3) {
            v[i] += shift3[i];
        }
    }

}
/// <summary>
/// ReLu activation function
/// </summary>
/// <param name="v">Unactivated layer</param>
void plyrANN::ReLu(double v[50]) {
    for (int i = 0; i < 25; i++) {
        if (v[i] < 0)
            v[i] = 0;
    }
}
/// <summary>
/// Evaluates a board position (UPDATES V).
/// </summary>
/// <param name="v">Boardstate</param>
/// <returns>Value estimation</returns>
double plyrANN::EvalFcn(double v[50]) {
    if (v == nullptr) {
        return 0;
    }
    
    matrixMultiply(v, 1);
    shift(v, 1);
    ReLu(v);
    matrixMultiply(v, 2);
    shift(v, 2);
    ReLu(v);
    matrixMultiply(v, 3);
    shift(v, 3);
    ReLu(v);
    return lastLayerMultiply(v);
}
/// <summary>
/// Takes a turn using ANN evaluation.
/// </summary>
/// <returns></returns>
int plyrANN::takeTurn() {
    //cout << "IN ANN TAKE TURN" << endl;
    //getWorkers for workers
    int workers[NUM_WORKERS_PER_PLAYER];
    getWorkers(workers);
    //getTransitions for bool array
    bool moves[NUM_WORKERS_PER_PLAYER][8][8];
    getTransitions(moves);
    //extractNewBoardState for respective boardstates

    int valid = 0;

    double* boardstates[NUM_WORKERS_PER_PLAYER][8][8];
    for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
        for (int m = 0; m < 8; m++) {
            for (int p = 0; p < 8; p++) {
                if (moves[w][m][p] == 1) { //there is a move
                    double* v = new double[50];
                    extractNewBoardState(workers[w], m, p, getColor(), v);
                    boardstates[w][m][p] = v;
                    valid++;
                }
                else {
                    boardstates[w][m][p] = nullptr;
                }

            }
        }
    }

    //cout << "test1 " << valid << endl;
    if (valid == 0) {
        return -1;
    }


    //new array to fill up with board evaluations
    double eval[NUM_WORKERS_PER_PLAYER][8][8];
    for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
        for (int m = 0; m < 8; m++) {
            for (int p = 0; p < 8; p++) {
                eval[w][m][p] = EvalFcn(boardstates[w][m][p]);
                //cout << w << " " << m << " " << p << " " << eval[w][m][p] << endl;
            }
        }
    }

    //cout << "test2 " << endl;

    double max = numeric_limits<double>::lowest();
    //find max
    for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
        for (int m = 0; m < 8; m++) {
            for (int p = 0; p < 8; p++) {
                if (eval[w][m][p] > max && moves[w][m][p] == 1)
                    max = eval[w][m][p];
            }
        }
    }
    //cout << "test3 " << max << endl;


    threeIndex turns[128];
    int num_turns = 0;

    //populate array with moves within threshold of max
    for (int w = 0; w < NUM_WORKERS_PER_PLAYER; w++) {
        for (int m = 0; m < 8; m++) {
            for (int p = 0; p < 8; p++) {
                if (eval[w][m][p] >= max - threshold && moves[w][m][p] == 1) {
                    turns[num_turns].x = w;
                    turns[num_turns].y = m;
                    turns[num_turns].z = p;
                    num_turns++;
                }
            }
        }
    }
    //cout << "test4 " << num_turns << endl;


    int make_turn = rand() % num_turns;
    W[workers[turns[make_turn].x]].makeMove(turns[make_turn].y);
    W[workers[turns[make_turn].x]].makePlacement(turns[make_turn].z);


    //cout << "test5" << endl;

    if (Board[W[workers[turns[make_turn].x]].x][W[workers[turns[make_turn].x]].y] == 3)
        return 1;
    return 0;
}
/// <summary>
/// Place a worker on Board using random selection.
/// </summary>
/// <param name="index">Index of worker in W.</param>
void plyrANN::setWorker(int index) {
    int x, y;
    bool check = true;
    while (check) { //gets random coordinates until unoccupied space is found.
        check = false;
        x = rand() % BOARD_SIZE;
        y = rand() % BOARD_SIZE;
        for (int i = 0; i < NUM_WORKERS; i++) {
            if (W[i].x == x && W[i].y == y) {
                check = true;
                break;
            }
        }
    }
    W[index].x = x;
    W[index].y = y;
    W[index].color = getColor();
}

/*
Functions used for retrieving a particular value in the ANN for training purposes.
*/
double plyrANN::getEval(double x[50]) {
    double v[50];
    for (int i = 0; i < 50; i++) {
        v[i] = x[i];
    }

    matrixMultiply(v, 1);
    shift(v, 1);
    ReLu(v);
    matrixMultiply(v, 2);
    shift(v, 2);
    ReLu(v);
    matrixMultiply(v, 3);
    shift(v, 3);
    ReLu(v);
    return lastLayerMultiply(v);

}
double plyrANN::getWeight(int layer, int i, int j) {
    if (layer == 1) {
        return layer1[i][j];
    }
    else if (layer == 2) {
        return layer2[i][j];
    }
    else if (layer == 3) {
        return layer3[i][j];
    }
    else if (layer == 4){
        return layer4[1][j];
    }

    cout << "getWeight out of bounds" << endl;
    return 0;
}
double plyrANN::getBias(int layer, int i) {
    if (layer == 1) {
        return shift1[i];
    }
    else if (layer == 2) {
        return shift2[i];
    }
    else if (layer == 3) {
        return shift3[i];
    }

    cout << "getBias out of bounds" << endl;
    return 0;
}
double plyrANN::getLogit(const double x[50], int layer, int i) {
    double v[50];
    for (int i = 0; i < 50; i++) {
        v[i] = x[i];
    }

    matrixMultiply(v, 1);
    shift(v, 1);
    if (layer == 1) {
        return v[i];
    }
    ReLu(v);
    matrixMultiply(v, 2);
    shift(v, 2);
    if (layer == 2) {
        return v[i];
    }
    ReLu(v);
    matrixMultiply(v, 3);
    shift(v, 3);
    if (layer == 3) {
        return v[i];
    }

    cout << "getLogit out of bounds" << endl;
    return 0;
}
double plyrANN::getHidden(const double x[50], int layer, int i) {
    double v[50];
    for (int i = 0; i < 50; i++) {
        v[i] = x[i];
    }
    
    matrixMultiply(v, 1);
    shift(v, 1);
    ReLu(v);
    if (layer == 1) {
        return v[i];
    }
    matrixMultiply(v, 2);
    shift(v, 2);
    ReLu(v);
    if (layer == 2) {
        return v[i];
    }
    matrixMultiply(v, 3);
    shift(v, 3);
    ReLu(v);
    if (layer == 3) {
        return v[i];
    }

    cout << "getHidden out of bounds" << endl;
    return 0;
}

/*
Functions for adjusting and saving a ANN for training purposes.
*/
void plyrANN::updateANN(double g[7850]) {
    // Update W(4)
    for (int i = 0; i <= 49; i++) {
        if (i < 25 && layer4[0][i] != 0)
            layer4[0][i] += g[i];
    }

    // Update W(3)
    for (int i = 100; i <= 2599; i++) {
        int x = (i - 100) / 50;
        int y = (i - 100) % 50;
        if (x < 25 && layer3[x][y] != 0)
            layer3[x][y] += g[i];
    }

    // Update b(3)
    for (int i = 2600; i <= 2649; i++) {
        if(i - 2600 < 25)
            shift3[i - 2600] += g[i];
    }

    // Update W(2)
    for (int i = 2700; i <= 5199; i++) {
        int x = (i - 2700) / 50;
        int y = (i - 2700) % 50;
        if (x < 25 && layer2[x][y] != 0)
            layer2[x][y] += g[i];
    }

    // Update b(2)
    for (int i = 5200; i <= 5249; i++) {
        if(i - 5200 < 25)
            shift2[i - 5200] += g[i];
    }

    // Update W(1)
    for (int i = 5300; i <= 7799; i++) {
        int x = (i - 5300) / 50;
        int y = (i - 5300) % 50;
        if (x < 25 && layer1[x][y] != 0)
            layer1[x][y] += g[i];
    }

    // Update b(1)
    for (int i = 7800; i <= 7849; i++) {
        if(i - 7800 < 25)
            shift1[i - 7800] += g[i];
    }
}
void plyrANN::exportANN(string filename) {
    ofstream fout;
    fout.open(filename);

    if (!fout) {
        cerr << "exportANN(): Could not open file " << filename << " for output." << endl;
        exit(EXIT_FAILURE);
    }

    //layer1 weights
    for (int i = 0; i < 50; i++) {
        for (int j = 0; j < 50; j++) {
            fout << layer1[i][j] << " ";
        }
        fout << endl;  // Optionally add a newline for better readability in the file
    }

    //shift1 biases
    for (int i = 0; i < 50; i++) {
        fout << shift1[i] << " ";
    }
    fout << endl;

    //layer2 weights
    for (int i = 0; i < 50; i++) {
        for (int j = 0; j < 50; j++) {
            fout << layer2[i][j] << " ";
        }
        fout << endl;
    }

    //shift2 biases
    for (int i = 0; i < 50; i++) {
        fout << shift2[i] << " ";
    }
    fout << endl;

    //layer3 weights
    for (int i = 0; i < 50; i++) {
        for (int j = 0; j < 50; j++) {
            fout << layer3[i][j] << " ";
        }
        fout << endl;
    }
    //shift3 biases
    for (int i = 0; i < 50; i++) {
        fout << shift3[i] << " ";
    }
    fout << endl;

    //layer4 weights
    for (int j = 0; j < 50; j++) {
        fout << layer4[0][j] << " ";
    }
    fout << endl;

    fout.close();
    cout << "Successfully exported to file: " << filename << endl;
}



//MINIMAX FUNCTIONS
//populates tree with a bool whether or not a move is valid
void validGameTree(double v[50], double tree[128]) {
    for (int i = 0; i < 128; i++) {
        tree[i] = 0;
    }

    int workers[2];
    int index = 0;

    for (int i = 25; i < 50; i++) {
        if (v[i] == 1) {
            workers[index] = i;
            index++;
        }
    }

    double temp[50];
    int x;
    int y;
    int newindex;

    for (int w = 0; w <= 1; w++) {
        for (int m = 0; m < 8; m++) {

            //copy v into temp
            for (int i = 0; i < 50; i++) {
                temp[i] = v[i];
            }

            if (checkValidMove(m,workers[w],temp)) { //if there is a valid move
                temp[workers[w]] = 0; //remove the worker from boardstate

                x = (workers[w] - 25) / 5; //find the workers x coordinate
                y = (workers[w] - 25) % 5; //find the workers y coordinate
                adjacentSpace(m, x, y); //adjust them both by move
                newindex = (x * 5) + y + 25; //find the index of the new worker
                temp[newindex] = 1; //place worker in new index

                for (int p = 0; p < 8; p++) {
                    if (checkValidPlacement(p, newindex, temp)) {
                        tree[(w*64) + (m * 8) + p] = 1;
                    }
                    else {
                        tree[(w * 64) + (m * 8) + p] = 0;

                    }
                }
            }
            else { //not a valid move
                for (int p = 0; p < 8; p++) { //therefore none of the placements are valid.
                    tree[(w * 64) + (m * 8) + p] = 0;
                }
            }
        }
    }


}

//checks to see if a move from a worker is valid given a boardstate
bool checkValidMove(int move, int index, double v[50]) {
    //index = worker
    int pos = index - 25;
    int x = pos / 5;
    int y = pos % 5;

    int x2 = x;
    int y2 = y;

    if (!adjacentSpace(move, x2, y2)) { //out of bounds
        return false;
    }

    //check occupied
    int pos2 = (x2 * 5) + y2 + 25;
    if (v[pos2] != 0) { //occupied space
        return false;
    }
    
    //check height difference
    if (v[pos2] > v[pos] + 1) { //more than 1 increase in height
        return false;
    }

    //check complete tower
    if (v[pos2] == 4) {
        return false;
    }
    return true;
}

//checks to see if a placement from a worker is valid given a boardstate
bool checkValidPlacement(int move, int index, double v[50]) {
    //index = worker
    int pos = index - 25;
    int x = pos / 5;
    int y = pos % 5;

    int x2 = x;
    int y2 = y;

    if (!adjacentSpace(move, x2, y2)) { //out of bounds
        return false;
    }

    //check occupied
    int pos2 = (x2 * 5) + y2 + 25;
    if (v[pos2] != 0) { //occupied space
        return false;
    }

    //check complete tower
    if (v[pos2] == 4) {
        return false;
    }
    return true;
}

//populates tree with boardstates that are valid
void buildGameTree(double v[50], double tree[128][50]) {
   
    for (int i = 0; i < 128; i++) {
        for (int j = 0; j < 50; j++) {
            tree[i][j] = -1;
        }
    }
    
    double valid[128];
    validGameTree(v, valid);

    int workers[2];
    int index = 0;

    for (int i = 25; i < 50; i++) {
        if (v[i] == 1) {
            workers[index] = i;
            index++;
        }
    }


    double temp[50];
    int mx;
    int my;
    int px;
    int py;

    
    for (int w = 0; w < 2; w++) {
        for (int m = 0; m < 8; m++) {
            for (int i = 0; i < 50; i++) {
                temp[i] = v[i];
            }
            mx = (workers[w] - 25) / 5;
            my = (workers[w] - 25) % 5;
            temp[workers[w]] = 0;
            adjacentSpace(m, mx, my);
            temp[(mx * 5) + my + 25] = 1;

            for (int p = 0; p < 8; p++) {
                if (valid[(w * 64) + (m * 8) + p] == 1) {
                    px = mx;
                    py = my;
                    adjacentSpace(p, px, py);
                    for (int i = 0; i < 25; i++) {
                        if (i == (px * 5) + py) {
                            tree[(w * 64) + (m * 8) + p][i] = temp[i]+1;
                        }
                        else {
                            tree[(w * 64) + (m * 8) + p][i] = temp[i];

                        }
                    }
                    for (int i = 25; i < 50; i++) {
                        if (temp[i] == 1) {
                            tree[(w * 64) + (m * 8) + p][i] = -1;
                        }
                        else if (temp[i] == -1) {
                            tree[(w * 64) + (m * 8) + p][i] = 1;
                        }
                        else {
                            tree[(w * 64) + (m * 8) + p][i] = 0;
                        }
                    }


                }
            }
        }
    }
}

//a boolean value to interpret a boardstate as valid or not
bool isValidState(double board[50]) {
    if (board[0] == -1) {
        return false;
    }
    return true;
}

//minimax that invokes a ANN evaluation fcn
double minimax(int depth, double board[50], bool isMaximizingPlayer, int maxDepth, plyrANN* p) {
    double tree[128][50];
    buildGameTree(board, tree);

    // Base case: maximum depth is reached or no further moves
    if (depth == maxDepth) {
        return p->getEval(board);
    }

    double bestValue;
    if (isMaximizingPlayer) {
        bestValue = -INFINITY;
        for (int i = 0; i < 128; i++) {
            if (!isValidState(tree[i])) continue; // Check if the state is valid
            double value = minimax(depth + 1, tree[i], false, maxDepth, p);
            bestValue = max(bestValue, value);
        }
    }
    else {
        bestValue = INFINITY;
        for (int i = 0; i < 128; i++) {
            if (!isValidState(tree[i])) continue; // Check if the state is valid
            double value = minimax(depth + 1, tree[i], true, maxDepth, p);
            bestValue = min(bestValue, value);
        }
    }
    return bestValue;
}

//minimax that invokes a heuristic evaluation fcn
double minimax(int depth, double board[50], bool isMaximizingPlayer, int maxDepth, plyrHeuristic* p) {
    double tree[128][50];
    buildGameTree(board, tree);

    // Base case: maximum depth is reached or no further moves
    if (depth == maxDepth) {
        return p->getEval(board);
    }

    double bestValue;
    if (isMaximizingPlayer) {
        bestValue = -INFINITY;
        for (int i = 0; i < 128; i++) {
            if (!isValidState(tree[i])) continue; // Check if the state is valid
            double value = minimax(depth + 1, tree[i], false, maxDepth, p);
            bestValue = max(bestValue, value);
        }
    }
    else {
        bestValue = INFINITY;
        for (int i = 0; i < 128; i++) {
            if (!isValidState(tree[i])) continue; // Check if the state is valid
            double value = minimax(depth + 1, tree[i], true, maxDepth, p);
            bestValue = min(bestValue, value);
        }
    }
    return bestValue;
}