#include "config.h"
#include "player.h"

using namespace std;

void initWorkers(Player* p1, Player* p2, bool print = true);
void buildANN(string filename);
int searchStates(string S[2000], int size, string p);
string trim(const string& s);
double WinLoseState(double V[50]);


int Board[BOARD_SIZE][BOARD_SIZE];
Worker W[NUM_WORKERS];



int main() {
	//step to start on
	char start = 'C';

	//step to terminate on (if range is true)
	char end = 'C';

	//whether or not to do multiple steps
	//end = start;

	//What ANN to train from...
	string ANNFile = "ANNL.txt";
	//What ANN to output to
	string ANNOutput = "newANN.txt";



	/*
	* This segment plays 10 games and stores non-repeating moves in saveboards.txt
	*/
	if (start == 'A') {
		//play 10 games, record boardstates

		string boardstates[2000];
		string current;

		int numstates = 0;

		srand(time(NULL));

		Player* p1;
		Player* p2;

		for (int i = 0; i < 10; i++) {

			p1 = new plyrHeuristic("Player 1", RED);
			p2 = new plyrHeuristic("Player 2", GREEN);

			initWorkers(p1, p2, false);

			int result = 0;
			int turn = 0;
			bool gameLoop = true;
			int num_turns = 0;

			while (gameLoop) {
				num_turns++;
				turn = (turn + 1) % 2;
				//printBoard();

				if (turn == 0) {
					exportBoardState(p1->getColor(), current);
					//cout << current << endl;
					if (searchStates(boardstates, numstates, current) == -1) {
						boardstates[numstates] = current;
						numstates++;
					}

					result = p1->takeTurn();
				}
				else {
					exportBoardState(p2->getColor(), current);
					//cout << current << endl;
					if (searchStates(boardstates, numstates, current) == -1) {
						boardstates[numstates] = current;
						numstates++;
					}

					result = p2->takeTurn();
				}

				if (result != 0) {
					exportBoardState(p2->getColor(), current);
					boardstates[numstates] = current;
					cout << numstates << endl;
					numstates++;
					printBoard();
					//cout << "Game Over!" << endl;
					break;
				}
			}

			delete p1;
			delete p2;
			for (int i = 0; i < BOARD_SIZE; i++) {
				for (int j = 0; j < BOARD_SIZE; j++) {
					Board[i][j] = 0;
				}
			}

		}

		ofstream fout;
		string filename = "saveboards.txt";
		fout.open(filename);
		if (!fout) {
			cerr << "main(): Could not open output file " << filename << endl;
			exit(EXIT_FAILURE);
		}

		for (int i = 0; i < numstates; i++) {
			fout << boardstates[i] << endl;
		}
		fout.close();

		cout << "DONE!" << endl;

	}

	/*
	* This segment finds measured winrates for each board within saveboards.txt
	*/
	if (start <= 'B' && end >= 'B') {
		//measure values for stored boardstates

		Player* p1;
		Player* p2;
		p1 = new plyrHeuristic("Player 1", RED);
		p2 = new plyrHeuristic("Player 2", GREEN);

		int original_color = 0;
		double original_board[50];

		double boardstates[2000][50];
		double measuredvalues[2000];
		int numstates = 0;

		/*
		Read in each boardstate from saveboards.txt to boardstates[2000][50]
		*/
		ifstream fin;
		string filename = "saveboards.txt";
		fin.open(filename);
		if (!fin.is_open()) {
			cerr << "main(): Could not open file " << filename << endl;
			exit(EXIT_FAILURE);
		}
		while (!fin.eof()) {
			importBoardState(original_color, p1, p2, fin);
			if (original_color == 0) {
				extractBoardState(p1->getColor(), original_board);
			}
			else {
				extractBoardState(p2->getColor(), original_board);

			}

			for (int i = 0; i < 50; i++) {
				boardstates[numstates][i] = original_board[i];
			}
			//printVector(boardstates[numstates]);
			numstates++;
		}

		fin.close();
		cout << "NUMSTATES: " << numstates << endl << endl;

		/*
		Iterate across all boardstates held within
		*/
		for (int index = 0; index < numstates; index++) {
			int turn = p1->getColor();
			importBoardState(turn, p1, p2, boardstates[index]);


			double p1_wins = 0.0;
			double p2_wins = 0.0;

			int n = 1000;
			int check = WinLoseState(boardstates[index]);

			if (check == 1) {
				p1_wins = n;
				setColor(RED);
			}
			else if (check == -1) {
				p2_wins = n;
				setColor(GREEN);
			}
			cout << "Boardstate " << index << ":" << endl;
			printVector(boardstates[index]);
			//printBoard();
			if (check == 0) {
				for (int i = 0; i < n; i++) {
					int turn = p1->getColor();
					importBoardState(turn, p1, p2, boardstates[index]);
					int result = 0;
					bool gameLoop = true;
					while (gameLoop) {
						turn = (turn + 1) % 2;
						//printBoard();

						if (turn == 0) {
							result = p1->takeTurn();
						}
						else {
							result = p2->takeTurn();
						}

						if (result != 0) {
							//printBoard();
							if (result == 1) {
								if (turn == 0) {
									p1_wins++;
								}
								else
									p2_wins++;
							}
							else {
								if (turn == 0) {
									p2_wins++;
								}
								else
									p1_wins++;
							}

							break;
						}
					}
				}
			}

			cout << "P1wins: " << p1_wins << endl;
			cout << "P2wins: " << p2_wins << endl;
			measuredvalues[index] = (p1_wins - p2_wins) / n;
			cout << "Value: " << setprecision(4) << measuredvalues[index] << endl;
			setColor(WHITE);
			cout << endl;

		}

		ofstream fout;
		filename = "savevalues.txt";
		fout.open(filename);
		if (!fout) {
			cerr << "main(): Could not open output file " << filename << endl;
			exit(EXIT_FAILURE);
		}
		for (int i = 0; i < numstates; i++) {
			fout << measuredvalues[i] << endl;
		}
		fout.close();
	}

	/*
	* This segment measures gradient across each variable and outputs a new ANN
	*/
	if (start <= 'C' && end >= 'C') {
		ifstream fin;
		string filename = "saveboards.txt";
		fin.open(filename);
		if (!fin.is_open()) {
			cerr << "main(): Could not open file " << filename << endl;
			exit(EXIT_FAILURE);
		}


		Player* p1;
		Player* p2;
		p1 = new plyrHeuristic("Player 1", RED);
		p2 = new plyrHeuristic("Player 2", GREEN);

		int original_color = 0;
		double board[50];

		double boardstates[2000][50];
		double measuredvalues[2000];
		int numstates = 0;
		int numvalues = 0;

		while (!fin.eof()) {
			importBoardState(original_color, p1, p2, fin);
			if (original_color == 0) {
				extractBoardState(p1->getColor(), board);
			}
			else {
				extractBoardState(p2->getColor(), board);

			}

			for (int i = 0; i < 50; i++) {
				boardstates[numstates][i] = board[i];
			}
			//printVector(boardstates[numstates]);
			numstates++;
		}
		fin.close();

		double temp;
		filename = "savevalues.txt";
		fin.open(filename);
		if (!fin.is_open()) {
			cerr << "main(): Could not open file " << filename << endl;
			exit(EXIT_FAILURE);
		}
		while (!fin.eof()) {
			fin >> measuredvalues[numvalues];
			numvalues++;
		}
		fin.close();


		// Calculate gradient only for the indices used in updateANN
		double full_gradient[7850] = {};

		double** gradient;
		gradient = new double* [900];
		for (int i = 0; i < numstates; i++) {
			gradient[i] = new double[7850];
			for (int j = 0; j < 7850; j++) {
				gradient[i][j] = 0;
			}
		}
		cout << "beginning training" << endl;
	

		plyrANN* p3;
		p3 = new plyrANN(ANNFile, "Player 1", RED);

		double sum = 0;
		double hold = 0;

		//LAYER 4 --- W(4) and h(3)
		for (int i = 0; i < 50; i++) {
			for (int state = 0; state < numstates; state++) {
				gradient[state][i] = p3->getHidden(boardstates[state], 3, i);
				gradient[state][i + 50] = p3->getWeight(4, 0, i); //dh(3)
			}
		}

		cout << "LAYER4 W(4) and H(3) DONE" << endl;

		//LAYER 3 --- W(3)
		for (int i = 0; i < 50; i++) {
			for (int j = 0; j < 50; j++) {
				for (int state = 0; state < numstates; state++) {
					if (p3->getLogit(boardstates[state], 3, i) > 0) { //check if z(3)_i > 0
						gradient[state][100 + (i * 50) + j] = gradient[state][i + 50] * p3->getHidden(boardstates[state], 2, j);
					}
					else {
						gradient[state][100 + (i * 50) + j] = 0;
					}
				}
			}
		}
		cout << "LAYER3 W(3) DONE" << endl;
		//b(3)
		for (int i = 0; i < 50; i++) {
			for (int state = 0; state < numstates; state++) {
				if (p3->getLogit(boardstates[state], 3, i) > 0) { //check if z(3)_i > 0
					gradient[state][2600 + i] = gradient[state][i + 50];
				}
				else {
					gradient[state][2600 + i] = 0;
				}
			}
		}
		cout << "LAYER3 B(3) DONE" << endl;
		//h(2)
		for (int i = 0; i < 50; i++) {
			for (int state = 0; state < numstates; state++) {
				sum = 0;
				hold = 0;
				for (int j = 0; j < 50; j++) {
					hold = p3->getWeight(3, j, i);
					if (hold != 0 && p3->getLogit(boardstates[state], 3, j) > 0) { //check if z(3)_j > 0
						sum += hold * gradient[state][j + 50];
					}
				}
				gradient[state][2650 + i] = sum;
			}
		}
		cout << "LAYER3 H(2) DONE" << endl;

		//LAYER 2 --- W(2)
		for (int i = 0; i < 50; i++) {
			for (int j = 0; j < 50; j++) {
				for (int state = 0; state < numstates; state++) {
					if (p3->getLogit(boardstates[state], 2, i) > 0) { //check if z(2)_i > 0
						gradient[state][2700 + (i * 50) + j] = gradient[state][i + 2650] * p3->getHidden(boardstates[state], 1, j);
					}
					else {
						gradient[state][2700 + (i * 50) + j] = 0;
					}
				}
			}
		}
		cout << "LAYER2 W(2) DONE" << endl;
		//b(2)
		for (int i = 0; i < 50; i++) {
			for (int state = 0; state < numstates; state++) {
				if (p3->getLogit(boardstates[state], 2, i) > 0) { //check if z(2)_i > 0
					gradient[state][5200 + i] = gradient[state][i + 2650];
				}
				else {
					gradient[state][5200 + i] = 0;
				}
			}
		}
		cout << "LAYER2 B(2) DONE" << endl;
		//h(1)
		for (int i = 0; i < 50; i++) {
			for (int state = 0; state < numstates; state++) {
				sum = 0;
				hold = 0;
				for (int j = 0; j < 50; j++) {
					hold = p3->getWeight(2, j, i);
					if (hold != 0 && p3->getLogit(boardstates[state], 2, j) > 0) { //check if z(2)i > 0
						sum += hold * gradient[state][j + 2650];
					}
				}
				gradient[state][5250 + i] = sum;
			}
		}
		cout << "LAYER2 H(1) DONE" << endl;

		//LAYER 1 --- W(1)
		for (int i = 0; i < 50; i++) {
			for (int j = 0; j < 50; j++) {
				for (int state = 0; state < numstates; state++) {
					if (p3->getLogit(boardstates[state], 1, i) > 0) { //check if z(1)_i > 0
						gradient[state][5300 + (i * 50) + j] = gradient[state][i + 5250] * boardstates[state][j];
					}
					else {
						gradient[state][5300 + (i * 50) + j] = 0;
					}
				}
			}
		}
		cout << "LAYER1 W(1) DONE" << endl;

		//b(1)
		for (int i = 0; i < 50; i++) {
			for (int state = 0; state < numstates; state++) {
				if (p3->getLogit(boardstates[state], 1, i) > 0) { //check if z(1)_i > 0
					gradient[state][7800 + i] = gradient[state][i + 5250];
				}
				else {
					gradient[state][7800 + i] = 0;
				}
			}
		}
		cout << "LAYER1 B(1) DONE" << endl;

		cout << "Calculating full gradient..." << endl;

		// Layer 1 Weights and Biases (Indices 5300 to 7849 for W(1) and b(1))
		for (int i = 5300; i <= 7849; i++) {
			sum = 0;
			for (int state = 0; state < numstates; state++) {
				double contribution = (p3->getEval(boardstates[state]) - measuredvalues[state]) * gradient[state][i];
				sum += contribution;
			}
			full_gradient[i] = -(sum / numstates);
		}

		// Layer 2 Weights and Biases (Indices 2700 to 5249 for W(2) and b(2))
		for (int i = 2700; i <= 5249; i++) {
			sum = 0;
			for (int state = 0; state < numstates; state++) {
				double contribution = (p3->getEval(boardstates[state]) - measuredvalues[state]) * gradient[state][i];
				sum += contribution;
			}
			full_gradient[i] = -(sum / numstates);
		}

		// Layer 3 Weights and Biases (Indices 100 to 2649 for W(3) and b(3))
		for (int i = 100; i <= 2649; i++) {
			sum = 0;
			for (int state = 0; state < numstates; state++) {
				double contribution = (p3->getEval(boardstates[state]) - measuredvalues[state]) * gradient[state][i];
				sum += contribution;
			}
			full_gradient[i] = -(sum / numstates);
		}

		// Layer 4 Weights (Indices 0 to 49 for W(4))
		for (int i = 0; i <= 49; i++) {
			sum = 0;
			for (int state = 0; state < numstates; state++) {
				double contribution = (p3->getEval(boardstates[state]) - measuredvalues[state]) * gradient[state][i];
				sum += contribution;
			}
			full_gradient[i] = -(sum / numstates);
		}

		for (int i = 0; i < 500; i++) {
			delete[] gradient[i];
		}
		delete[] gradient;

		//apply learning rate
		for (int i = 0; i < 7850; i++) {
			full_gradient[i] = full_gradient[i] * 0.001;
			cout << "[" << i << "]: " << full_gradient[i] << endl;
		}


		p3->updateANN(full_gradient);
		p3->exportANN(ANNOutput);
	}

	/*
	* This segment tests two different ANN outputs
	*/
	if (start <= 'D' && end >= 'D') {
		cout << "TEST" << endl;
		ifstream fin;
		string filename = "saveboards.txt";
		fin.open(filename);
		if (!fin.is_open()) {
			cerr << "main(): Could not open file " << filename << endl;
			exit(EXIT_FAILURE);
		}
		
		
		plyrHeuristic* p1;
		p1 = new plyrHeuristic("HEURSTIC", RED);
		Player* p2;
		p2 = new plyrHeuristic("Player 2", GREEN);

		plyrANN* p3;
		p3 = new plyrANN("ANN0.txt", "ORIG ANN", RED);
		plyrANN* p4;
		p4 = new plyrANN(ANNOutput, "NEW ANN", RED);


		int original_color = 0;
		double board[50];

		double boardstates[2000][50];
		double measuredvalues[2000];
		int numstates = 0;
		int numvalues = 0;

		while (!fin.eof()) {
			importBoardState(original_color, p1, p2, fin);
			if (original_color == 0) {
				extractBoardState(p1->getColor(), board);
			}
			else {
				extractBoardState(p2->getColor(), board);

			}

			for (int i = 0; i < 50; i++) {
				boardstates[numstates][i] = board[i];
			}
			//printVector(boardstates[numstates]);
			numstates++;
		}
		fin.close();

		double temp;
		filename = "savevalues.txt";
		fin.open(filename);
		if (!fin.is_open()) {
			cerr << "main(): Could not open file " << filename << endl;
			exit(EXIT_FAILURE);
		}
		while (!fin.eof()) {
			fin >> measuredvalues[numvalues];
			numvalues++;
		}
		fin.close();

		for (int i = 0; i < numstates; i++) {
			cout << "BOARDSTATE " << i << endl;
			printVector(boardstates[i]);
			cout << "MEASURED: " << measuredvalues[i] << endl;
			cout << p1->getName() << ": " << p1->getEval(boardstates[i]) << endl;
			cout << p3->getName() << ": " << p3->getEval(boardstates[i]) << endl;
			cout << p4->getName() << ": " << p4->getEval(boardstates[i]) << endl;
			cout << endl;
		}

	}


	return 0;
}

/// <summary>
/// Initializes all 4 workers at game start.
/// </summary>
/// <param name="p1">Player 1</param>
/// <param name="p2">Player 2</param>
/// <param name="print">Whether or not to printboard between selections</param>
void initWorkers(Player* p1, Player* p2, bool print) {
	int turn = rand() % 2;

	for (int i = 0; i < NUM_WORKERS; i++) {
		if (print == true)
			printBoard();
		turn = (turn + 1) % 2;
		if (turn == 0) {
			p1->setWorker(i);
		}
		else {
			p2->setWorker(i);
		}
	}
	return;
}
/// <summary>
/// Builds the most basic ANN structure
/// </summary>
/// <param name="filename">Filename for output</param>
void buildANN(string filename) {
	ofstream fout;
	fout.open(filename);

	//layer 1
	for (int i = 0; i < 50; i++) {
		if (i < 25) {
			for (int j = 0; j < 25; j++) {
				if (i == j || (i == j + 1 && j % 5 != 4) || (i == j - 1 && j % 5 != 0)) {
					fout << 1 << " ";
				}
				else if (i == j - 5 || (i == j - 5 + 1 && j % 5 != 4) || (i == j - 5 - 1 && j % 5 != 0)) {
					fout << 1 << " ";
				}
				else if (i == j + 5 || (i == j + 5 + 1 && j % 5 != 4) || (i == j + 5 - 1 && j % 5 != 0)) {
					fout << 1 << " ";
				}
				else {
					fout << 0 << " ";
				}
			}
			for (int j = 25; j < 50; j++) {
				if (i + 25 == j || (i + 25 == j + 1 && j % 5 != 4) || (i + 25 == j - 1 && j % 5 != 0)) {
					fout << 5 << " ";
				}
				else if (i + 25 == j - 5 || (i + 25 == j - 5 + 1 && j % 5 != 4) || (i + 25 == j - 5 - 1 && j % 5 != 0)) {
					fout << 5 << " ";
				}
				else if (i + 25 == j + 5 || (i + 25 == j + 5 + 1 && j % 5 != 4) || (i + 25 == j + 5 - 1 && j % 5 != 0)) {
					fout << 5 << " ";
				}
				else {
					fout << 0 << " ";
				}
			}

		}
		else { //identity matrix for worker positions
			for (int j = 0; j < 50; j++) {
				if (i == j) {
					fout << 1 << " ";
				}
				else
					fout << 0 << " ";
			}
		}
		fout << endl;
	}

	//shift 1
	for (int i = 0; i < 25; i++) {
		fout << -2 << " ";
	}
	for (int i = 0; i < 25; i++) {
		fout << 0 << " ";
	}
	fout << endl;

	//layer 2
	for (int i = 0; i < 50; i++) {
		if (i < 25) {
			for (int j = 0; j < 25; j++) {
				if (i == j || (i == j + 1 && j % 5 != 4) || (i == j - 1 && j % 5 != 0)) {
					fout << 1 << " ";
				}
				else if (i == j - 5 || (i == j - 5 + 1 && j % 5 != 4) || (i == j - 5 - 1 && j % 5 != 0)) {
					fout << 1 << " ";
				}
				else if (i == j + 5 || (i == j + 5 + 1 && j % 5 != 4) || (i == j + 5 - 1 && j % 5 != 0)) {
					fout << 1 << " ";
				}
				else {
					fout << 0 << " ";
				}
			}
			for (int j = 25; j < 50; j++) {
				if (i + 25 == j || (i + 25 == j + 1 && j % 5 != 4) || (i + 25 == j - 1 && j % 5 != 0)) {
					fout << 5 << " ";
				}
				else if (i + 25 == j - 5 || (i + 25 == j - 5 + 1 && j % 5 != 4) || (i + 25 == j - 5 - 1 && j % 5 != 0)) {
					fout << 5 << " ";
				}
				else if (i + 25 == j + 5 || (i + 25 == j + 5 + 1 && j % 5 != 4) || (i + 25 == j + 5 - 1 && j % 5 != 0)) {
					fout << 5 << " ";
				}
				else {
					fout << 0 << " ";
				}
			}

		}
		else { //identity matrix for worker positions
			for (int j = 0; j < 50; j++) {
				if (i == j) {
					fout << 1 << " ";
				}
				else
					fout << 0 << " ";
			}
		}
		fout << endl;
	}

	//shift 2
	for (int i = 0; i < 25; i++) {
		fout << 2 << " ";
	}
	for (int i = 0; i < 25; i++) {
		fout << 0 << " ";
	}
	fout << endl;

	//layer 3
	for (int i = 0; i < 50; i++) {
		if (i < 25) {
			for (int j = 0; j < 25; j++) {
				if (i == j || (i == j + 1 && j % 5 != 4) || (i == j - 1 && j % 5 != 0)) {
					fout << 0.25 << " ";
				}
				else if (i == j - 5 || (i == j - 5 + 1 && j % 5 != 4) || (i == j - 5 - 1 && j % 5 != 0)) {
					fout << 0.25 << " ";
				}
				else if (i == j + 5 || (i == j + 5 + 1 && j % 5 != 4) || (i == j + 5 - 1 && j % 5 != 0)) {
					fout << 0.25 << " ";
				}
				else {
					fout << 0 << " ";
				}
			}
			for (int j = 25; j < 50; j++) {
				if (i + 25 == j || (i + 25 == j + 1 && j % 5 != 4) || (i + 25 == j - 1 && j % 5 != 0)) {
					fout << 3 << " ";
				}
				else if (i + 25 == j - 5 || (i + 25 == j - 5 + 1 && j % 5 != 4) || (i + 25 == j - 5 - 1 && j % 5 != 0)) {
					fout << 3 << " ";
				}
				else if (i + 25 == j + 5 || (i + 25 == j + 5 + 1 && j % 5 != 4) || (i + 25 == j + 5 - 1 && j % 5 != 0)) {
					fout << 5 << " ";
				}
				else {
					fout << 0 << " ";
				}
			}

		}
		else { //identity matrix for worker positions
			for (int j = 0; j < 50; j++) {
				if (i == j) {
					fout << 1 << " ";
				}
				else
					fout << 0 << " ";
			}
		}
		fout << endl;
	}

	//shift 3
	for (int i = 0; i < 25; i++) {
		fout << 2 << " ";
	}
	for (int i = 0; i < 25; i++) {
		fout << 0 << " ";
	}
	fout << endl;

	//layer 4
	for (int i = 0; i < 25; i++) {
		fout << 0.1 << " ";
	}
	for (int i = 0; i < 25; i++) {
		fout << 0 << " ";
	}
	fout << endl;

	fout.close();
}
/// <summary>
/// Search array for target 
/// </summary>
/// <param name="S">Array to search</param>
/// <param name="size">Actual size of S</param>
/// <param name="p">Target</param>
/// <returns>index if found, -1 if not found</returns>
int searchStates(string S[2000], int size, string p) {
	for (int i = 0; i < size; i++) {
		if (p == S[i]) {
			return i;
		}
	}
	return -1;
}
//referenced stackoverflow... removes trailing whitespace from a string
string trim(const string& s) {
	size_t end = s.find_last_not_of("\t\n\r\f\v");
	return (end == string::npos) ? "" : s.substr(0, end + 1);
}
/// <summary>
/// Finds whether or not a state is already won or lost.
/// </summary>
/// <param name="V"></param>
/// <returns>1 for win, -1 for lose, 0 for neither</returns>
double WinLoseState(double V[50]) {
	for (int i = 0; i < 25; i++) {
		if (V[i] == 3 && V[i + 25] == 1)
			return 1;
		if (V[i] == 3 && V[i + 25] == -1)
			return -1;
	}
	return 0;
}

